//Baptista encryption based on the logistic map
module bptlogenc(
    input       clock,
    input       reset,

    output         plaintext_fifo_full,
    input          plaintext_fifo_open,
    input          plaintext_fifo_wren,
    input [31:0]   plaintext_fifo_data,

    output [31:0]  ciphertext_fifo_data,
    output         ciphertext_fifo_empty,
    output         ciphertext_fifo_eof,
    input          ciphertext_fifo_rden,
    input          ciphertext_fifo_open,

    //register interface
    input          reg_clk,   //register interface clock, asynchronous to the bus clock
    input          reg_wren,  //global write enable
    input [3:0]    reg_wstrb, //byte-level write enables
    input          reg_rden, //global read enable
    input [31:0]   reg_addr, //register address
    output reg [31:0] reg_rd_data,    //data read from core registers
    input [31:0]   reg_wr_data,    //data to be written to core registers
    output         reg_irq   //interrupt to the uprocessor

);


wire subtract_b_ready, subtract_result_ready, subtract_result_valid;
wire mul1_result_ready, mul1_result_valid;
wire mul2_result_ready, mul2_result_valid;
wire comp1_result_valid;
wire comp2_result_valid;
reg subtract_b_valid;

wire [31:0] one_minus_x;
wire [31:0] mul1_result;
wire [31:0] mul2_result;
wire [7:0] comp1_result;
wire [7:0] comp2_result;
wire [31:0] upper_limit;
wire [31:0] lower_limit;

reg [31:0] param;
reg [31:0] x;
reg [31:0] next_x;

wire mul2_within_range;

wire [7:0] plaintext_byte;
wire [15:0] ciphertext_value;

//input and output FIFOs
fifo_8x512 plaintext_fifo(
    .rst(reset | ~plaintext_fifo_open),
    .wr_clk(clock),
    .rd_clk(clock),
    .din(plaintext_fifo_data),
    .wr_en(plaintext_fifo_wren),
    .rd_en(plaintext_fifo_rden),
    .dout(plaintext_byte),
    .full(plaintext_fifo_full),
    .empty(plaintext_fifo_empty),
    .valid(plaintext_fifo_rdack)
);

fifo_32x512 ciphertext_fifo(
    .rst(reset | ~ciphertext_fifo_open),

    .wr_clk(clock),
    .wr_en(ciphertext_fifo_wren),
    .din({16'd0,ciphertext_value}),
    .full(ciphertext_fifo_full),

    .rd_clk(clock),
    .rd_en(ciphertext_fifo_rden),
    .dout(ciphertext_fifo_data),
    .empty(ciphertext_fifo_empty),
    .valid()
);

assign ciphertext_fifo_eof = 0;
assign reg_irq = 0;

//Lookup tables for upper and lower limits, and param register
bram_dp
#(
    .width(32),
    .depth_log(8)
)
upper_limit_lut
(
    //port A
    .clock_a(reg_clk),
    .enable_a((reg_wren | reg_rden) & (reg_addr[31:8] == 2)),
    .write_enable_a(reg_wren & (reg_addr[31:8] == 2)),
    .write_data_a(reg_wr_data),
    .addr_a(reg_addr[7:0]),
    .read_data_a(),
    //port B
    .clock_b(clock),
    .enable_b(1'b1),
    .write_enable_b(1'b0),
    .write_data_b(32'd0),
    .addr_b(plaintext_byte),
    .read_data_b(upper_limit)
);

bram_dp
#(
    .width(32),
    .depth_log(8)
)
lower_limit_lut
(
    //port A
    .clock_a(reg_clk),
    .enable_a((reg_wren | reg_rden) & (reg_addr[31:8] == 1)),
    .write_enable_a(reg_wren & (reg_addr[31:8] == 1)),
    .write_data_a(reg_wr_data),
    .addr_a(reg_addr[7:0]),
    .read_data_a(),
    //port B
    .clock_b(clock),
    .enable_b(1'b1),
    .write_enable_b(1'b0),
    .write_data_b(32'd0),
    .addr_b(plaintext_byte),
    .read_data_b(lower_limit)
);

always @(posedge clock)
    if(reg_wren & (reg_addr == 0))
        param <= reg_wr_data;

always @(posedge clock)
    if(reg_wren & (reg_addr == 1))
        x <= reg_wr_data;
    else if(comp1_result_valid & comp2_result_valid)
        x <= next_x;

//controller generates valids
controller ctrl(
    .clock(clock),
    .reset(reset),

    .read_plaintext(plaintext_fifo_rden),
    .plaintext_fifo_empty(plaintext_fifo_empty),

    .ciphertext_fifo_full(ciphertext_fifo_full),
    .iteration_done(comp1_result_valid & comp2_result_valid),
    .final_iteration(final_iteration),
    .write_ciphertext(ciphertext_fifo_wren),
    .ciphertext_found(mul2_within_range),
    .ciphertext_value(ciphertext_value)

);

always @(posedge clock)
    if(reset)
        subtract_b_valid <= 0;
    else 
        subtract_b_valid <= plaintext_fifo_rdack | (~final_iteration & ((comp1_result_valid & comp2_result_valid) & ~(mul2_within_range & (ciphertext_value >= 250))));
        
always @(posedge clock)
    next_x <= mul2_result;


//floating-point math for f(x) = lower_limit < (1-x)*x*param < upper_limit
fp_sub subtract(
  .aclk(clock),
  .aresetn(~reset),

  .s_axis_a_tready(),
  .s_axis_a_tvalid(subtract_b_valid),
  .s_axis_a_tdata(32'h3f800000),//IEEE754 single precision encoding for 1

  .s_axis_b_tready(subtract_b_ready),
  .s_axis_b_tvalid(subtract_b_valid),
  .s_axis_b_tdata(x),

  .m_axis_result_tready(subtract_result_ready),
  .m_axis_result_tvalid(subtract_result_valid),
  .m_axis_result_tdata(one_minus_x)
);

fp_mul mul_with_x(
  .aclk(clock),
  .aresetn(~reset),

  .s_axis_a_tready(),
  .s_axis_a_tvalid(subtract_result_valid),
  .s_axis_a_tdata(x),

  .s_axis_b_tready(subtract_result_ready),
  .s_axis_b_tvalid(subtract_result_valid),
  .s_axis_b_tdata(one_minus_x),

  .m_axis_result_tready(mul1_result_ready),
  .m_axis_result_tvalid(mul1_result_valid),
  .m_axis_result_tdata(mul1_result)
);

fp_mul mul_with_param(
  .aclk(clock),
  .aresetn(~reset),

  .s_axis_a_tready(),
  .s_axis_a_tvalid(mul1_result_valid),
  .s_axis_a_tdata(param),

  .s_axis_b_tready(mul1_result_ready),
  .s_axis_b_tvalid(mul1_result_valid),
  .s_axis_b_tdata(mul1_result),

  .m_axis_result_tready(mul2_result_ready),
  .m_axis_result_tvalid(mul2_result_valid),
  .m_axis_result_tdata(mul2_result)
);

fp_comp result_lt_upper(
  .aclk(clock),
  .aresetn(~reset),

  .s_axis_a_tready(mul2_result_ready),
  .s_axis_a_tvalid(mul2_result_valid),
  .s_axis_a_tdata(mul2_result),

  .s_axis_b_tready(),
  .s_axis_b_tvalid(mul2_result_valid),
  .s_axis_b_tdata(upper_limit),

  .m_axis_result_tready(1'b1),
  .m_axis_result_tvalid(comp1_result_valid),
  .m_axis_result_tdata(comp1_result)
);

fp_comp lower_lt_result(
  .aclk(clock),
  .aresetn(~reset),

  .s_axis_a_tready(),
  .s_axis_a_tvalid(mul2_result_valid),
  .s_axis_a_tdata(lower_limit),

  .s_axis_b_tready(mul2_result_ready),
  .s_axis_b_tvalid(mul2_result_valid),
  .s_axis_b_tdata(mul2_result),

  .m_axis_result_tready(1'b1),
  .m_axis_result_tvalid(comp2_result_valid),
  .m_axis_result_tdata(comp2_result)
);

assign mul2_within_range = (comp1_result == 1) & (comp2_result == 1);

endmodule
