module tb;

parameter LOWER_LIMIT_VAL = 0.2;//must be in range [0,1] and less than UPPER_LIMIT_VAL
parameter UPPER_LIMIT_VAL = 0.8;//must be in range [0,1] and greater than LOWER_LIMIT_VAL
parameter PARAM_VAL = 3.99; //must be in range (3.57,4]
parameter X_VAL = 0.3; //must be in range (0,1)

//file descriptors
integer plaintext_file, ciphertext_file, expected_ciphertext_file;
reg [31:0] plaintext_word;

real val;
reg [63:0] val_reg;

reg clock, reset;

reg program_done = 0;

reg reg_wren;
reg [31:0] reg_addr;
reg [31:0] reg_wr_data;

integer i;

reg plaintext_fifo_wren = 0;
reg [31:0] plaintext_fifo_data;
wire plaintext_fifo_full;

reg ciphertext_fifo_rdack = 0;
reg ciphertext_fifo_rden = 0;
wire ciphertext_fifo_empty;
wire [31:0] ciphertext_fifo_data;

reg [31:0] expected_ciphertext;

// Reset
initial begin
    reset = 1;
    // Wait 100 ns for global reset to finish
    #100 reset = 0;
end

// 100 MHz bus clock
initial begin
    clock = 0;
    forever #5 clock = ~clock;
end

//program the core LUTs and x/param registers
initial begin
    @(negedge reset);
    #1000
    //program x register
    @(negedge clock);
    reg_wren = 1;
    reg_addr = 1;
    val = X_VAL;
    val_reg = $realtobits(val);
    reg_wr_data = {val_reg[63],val_reg[62],val_reg[58:52],val_reg[51:29]};
    @(negedge clock);
    reg_wren = 0;
    reg_addr = 0;
    reg_wr_data = 0;
    //program param register
    @(negedge clock);
    reg_wren = 1;
    reg_addr = 0;
    val = PARAM_VAL;
    val_reg = $realtobits(val);
    reg_wr_data = {val_reg[63],val_reg[62],val_reg[58:52],val_reg[51:29]};
    @(negedge clock);
    reg_wren = 0;
    reg_addr = 0;
    reg_wr_data = 0;
    //program lower limit LUT
    for(i=0; i<256; i=i+1) begin
        @(negedge clock);
        reg_wren = 1;
        reg_addr = 256 + i;
        val = LOWER_LIMIT_VAL + ((UPPER_LIMIT_VAL - LOWER_LIMIT_VAL)*i/256);
        //crude conversion to single precision float in binary representation
        val_reg = $realtobits(val);
        reg_wr_data = {val_reg[63],val_reg[62],val_reg[58:52],val_reg[51:29]};
        @(negedge clock);
        reg_wren = 0;
        reg_addr = 0;
        reg_wr_data = 0;
    end
    //program upper limit LUT
    for(i=0; i<256; i=i+1) begin
        @(negedge clock);
        reg_wren = 1;
        reg_addr = 512 + i;
        val = LOWER_LIMIT_VAL + ((UPPER_LIMIT_VAL - LOWER_LIMIT_VAL)*(i+1)/256);
        //crude conversion to single precision float in binary representation
        val_reg = $realtobits(val);
        reg_wr_data = {val_reg[63],val_reg[62],val_reg[58:52],val_reg[51:29]};
        @(negedge clock);
        reg_wren = 0;
        reg_addr = 0;
        reg_wr_data = 0;
    end
    program_done = 1;
end

initial begin
    //open plaintext file for reading
    plaintext_file = $fopen("plain.txt","r");
    $display("Opened plaintext file");
    //open ciphertext file for writing
    ciphertext_file = $fopen("cipher.txt","wb");
    $display("Opened ciphertext file");
    //open expected ciphertext file for reading
    expected_ciphertext_file = $fopen("cipher.expected","r");
    $display("Opened expected ciphertext file");
    //wait for core programming to complete
    @(posedge program_done);
    #1000
    //write the plaintext into the core and read out ciphertext
    fork
        //plaintext process
        begin
            while(1) begin
                @(negedge clock)
                    plaintext_fifo_wren = 0;

                if(~plaintext_fifo_full) begin
                    if($fread(plaintext_word,plaintext_file) != 0) begin
                        plaintext_fifo_wren = 1;
                        plaintext_fifo_data = plaintext_word;
                    end
                end
            end
        end
        //ciphertext process
        begin
            while(1) begin
                @(negedge clock);
                ciphertext_fifo_rdack = ciphertext_fifo_rden;
                ciphertext_fifo_rden = ~ciphertext_fifo_empty;
                if(ciphertext_fifo_rdack) begin
                    $fwriteb(ciphertext_file,"%c%c",ciphertext_fifo_data[15:8],ciphertext_fifo_data[7:0]);
                    $fflush(ciphertext_file);
                    $fscanf(expected_ciphertext_file,"%d\n",expected_ciphertext);
                    if(expected_ciphertext !== ciphertext_fifo_data) begin
                        $display("Unexpected ciphertext (%d instead of %d), stopping simulation",ciphertext_fifo_data,expected_ciphertext);
                        $stop;
                    end else
                        $display("Ciphertext: %d",ciphertext_fifo_data);
                end
            end
        end
    join
end

bptlogenc dut(
    .clock(clock),
    .reset(reset),

    .plaintext_fifo_full(plaintext_fifo_full),
    .plaintext_fifo_open(1'b1),
    .plaintext_fifo_wren(plaintext_fifo_wren),
    .plaintext_fifo_data(plaintext_fifo_data),

    .ciphertext_fifo_data(ciphertext_fifo_data),
    .ciphertext_fifo_empty(ciphertext_fifo_empty),
    .ciphertext_fifo_eof(),
    .ciphertext_fifo_rden(ciphertext_fifo_rden),
    .ciphertext_fifo_open(1'b1),

    //register interface
    .reg_clk(clock),   //register interface clock, asynchronous to the bus clock
    .reg_wren(reg_wren),  //global write enable
    .reg_wstrb(4'b1111), //byte-level write enables
    .reg_rden(1'b0), //global read enable
    .reg_addr(reg_addr), //register address
    .reg_rd_data(),    //data read from core registers
    .reg_wr_data(reg_wr_data),    //data to be written to core registers
    .reg_irq()   //interrupt to the uprocessor

);


endmodule
